/*
 *   ALSA sequencer System services Client
 *   Copyright (c) 1998 by Frank van de Pol <F.K.W.van.de.Pol@inter.nl.net>
 *
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "driver.h"
#include "seq_system.h"
#include "seq_timer.h"
#include "seq_queue.h"

/* define TRACE_ANNOUNCEMENTSS to get every announcement traced to syslog/console */
#define TRACE_ANNOUNCEMENTS 1


/* internal client that provide system services, access to timer etc. */

/*
 * Port "Timer"
 *      - send tempo /start/stop etc. events to this port to manipulate the 
 *        queue's timer. The queue address is specified in the destination
 *        part of the address
 *      - this port supports subscription. The received timer events are 
 *        broadcasted to all subscribed clients.
 *
 * Port "Announce"
 *      - does not receive message
 *      - supports supscription. For each client or port attaching to or 
 *        detaching from the system an announcement is send to the subscribed
 *        clients.
 *
 * Idea: the subscription mechanism might also work handy for distributing 
 * synchronisation and timing information. In this case we would ideally have
 * a list of subscribers for each type of sync (time, tick), for each timing
 * queue.
 *
 */


/* client id of our system client */
static int sysclient = -1;

/* port id numbers for this client */
static int timer_port;
static int announce_port;



/* fill standard header data, source port & channel are filled in */
static void setheader(snd_seq_event_t * ev, int client, int port)
{
	ev->flags = SND_SEQ_EVENT_LENGTH_FIXED;

	ev->source.queue = -1;
	ev->source.client = sysclient;
	ev->source.port = announce_port;
	ev->source.channel = -1;

	ev->dest.queue = SND_SEQ_ADDRESS_SUBSCRIBERS;
	ev->dest.client = -1;
	ev->dest.port = -1;
	ev->dest.channel = -1;

	ev->flags = SND_SEQ_TIME_STAMP_REAL | SND_SEQ_TIME_MODE_ABS;
	ev->time.real.tv_sec = 0;
	ev->time.real.tv_nsec = 0;

	/* fill data */
	ev->data.addr.queue = -1;
	ev->data.addr.client = client;
	ev->data.addr.port = port;
	ev->data.addr.channel = -1;
}




/* entry points for broadcasting system events */
void snd_seq_system_client_ev_client_start(int client)
{
	snd_seq_event_t ev;
	
	setheader(&ev, client, -1);
	ev.type = SND_SEQ_EVENT_CLIENT_START;
	snd_seq_kernel_client_enqueue(sysclient, &ev);

#ifdef TRACE_ANNOUNCEMENTS
	snd_printk("seq: broadcast client %d start\n", client);
#endif
}

void snd_seq_system_client_ev_client_exit(int client)
{
	snd_seq_event_t ev;
	
	setheader(&ev, client, -1);
	ev.type = SND_SEQ_EVENT_CLIENT_EXIT;
	snd_seq_kernel_client_enqueue(sysclient, &ev);
	
#ifdef TRACE_ANNOUNCEMENTS
	snd_printk("seq: broadcast client %d exit\n", client);
#endif
}

void snd_seq_system_client_ev_client_change(int client)
{
	snd_seq_event_t ev;
	
	setheader(&ev, client, -1);
	ev.type = SND_SEQ_EVENT_CLIENT_CHANGE;
	snd_seq_kernel_client_enqueue(sysclient, &ev);
	
#ifdef TRACE_ANNOUNCEMENTS
	snd_printk("seq: broadcast client %d change\n", client);
#endif
}

void snd_seq_system_client_ev_port_start(int client, int port)
{
	snd_seq_event_t ev;
	
	setheader(&ev, client, port);
	ev.type = SND_SEQ_EVENT_PORT_START;
	snd_seq_kernel_client_enqueue(sysclient, &ev);
	
#ifdef TRACE_ANNOUNCEMENTS
	snd_printk("seq: broadcast client %d, port %d start\n", client, port);
#endif
}

void snd_seq_system_client_ev_port_exit(int client, int port)
{
	snd_seq_event_t ev;
	
	setheader(&ev, client, port);
	ev.type = SND_SEQ_EVENT_PORT_START;
	snd_seq_kernel_client_enqueue(sysclient, &ev);
	
#ifdef TRACE_ANNOUNCEMENTS
	snd_printk("seq: broadcast client %d, port %d start\n", client, port);
#endif
}

void snd_seq_system_client_ev_port_change(int client, int port)
{
	snd_seq_event_t ev;
	
	setheader(&ev, client, port);
	ev.type = SND_SEQ_EVENT_PORT_START;
	snd_seq_kernel_client_enqueue(sysclient, &ev);
	
#ifdef TRACE_ANNOUNCEMENTS
	snd_printk("seq: broadcast client %d, port %d start\n", client, port);
#endif
}




/* send timer event to all subscribed ports */
static void broadcast_timer_event(snd_seq_event_t *ev)
{
	ev->dest.queue = SND_SEQ_ADDRESS_SUBSCRIBERS;
	ev->dest.client = -1;
	ev->dest.port = -1;
	ev->dest.channel = -1;

	ev->flags = SND_SEQ_TIME_STAMP_REAL | SND_SEQ_TIME_MODE_ABS;
	ev->time.real.tv_sec = 0;
	ev->time.real.tv_nsec = 0;
			
	/* send event */
	snd_seq_kernel_client_enqueue(sysclient, ev);
}


/* FIXME: only broadcast timer event if the operation succeeded (if sender has permissions to change timer) */


/* call-back handler for timer events */
static int event_input_timer(snd_seq_event_t * ev, void *private_data)
{
	/* handle Timer events */
	switch (ev->type) {
			/* timer stuff */
		case SND_SEQ_EVENT_START:
			//snd_printk("start playing!\n");
			snd_seq_queue_timer_start(ev->dest.queue, ev->source.client);
			broadcast_timer_event(ev);
			break;

		case SND_SEQ_EVENT_CONTINUE:
			//snd_printk("continue playing!\n");
			snd_seq_queue_timer_continue(ev->dest.queue, ev->source.client);
			broadcast_timer_event(ev);
			break;

		case SND_SEQ_EVENT_STOP:
			//snd_printk("stop playing!\n");
			snd_seq_queue_timer_stop(ev->dest.queue, ev->source.client);
			broadcast_timer_event(ev);
			break;

		case SND_SEQ_EVENT_TEMPO:
			//snd_printk("tempo change!\n");
			snd_seq_queue_timer_set_tempo(ev->dest.queue, ev->source.client, ev->data.control.value);
			broadcast_timer_event(ev);
			break;
	}
	return 1;		/* success */
}


/* call-back function for event input */
static int event_input(snd_seq_event_t * ev, void *private_data)
{
	if (ev) {
		if (ev->dest.port == timer_port) {
			/* handle Timer events */
			event_input_timer(ev, private_data);
		}
	}
	return 1;		/* success */
}

/* register our internal client */
void snd_seq_system_client_init(void)
{

	snd_seq_client_callback_t callbacks;
	snd_seq_client_info_t inf;
	snd_seq_port_info_t port;

	/* register client */
	callbacks.input = event_input;
	sysclient = snd_seq_register_kernel_client(&callbacks, NULL);

	/* set our name */
	strcpy(inf.name, "System");
	snd_seq_kernel_client_ctl(sysclient, SND_SEQ_IOCTL_SET_CLIENT_INFO, &inf);

	/* register timer */
	strcpy(port.name, "Timer");
	port.capability = SND_SEQ_PORT_CAP_SUBSCRIPTION;
	port.port_type = 0;
	snd_seq_kernel_client_ctl(sysclient, SND_SEQ_IOCTL_CREATE_PORT, &port);
	timer_port = port.port;

	/* register announcement port */
	strcpy(port.name, "Announce");
	port.capability = SND_SEQ_PORT_CAP_SUBSCRIPTION;
	port.port_type = 0;
	snd_seq_kernel_client_ctl(sysclient, SND_SEQ_IOCTL_CREATE_PORT, &port);
	announce_port = port.port;
}


/* unregister our internal client */
void snd_seq_system_client_done(void)
{
	if (sysclient >= 0) {
		snd_seq_unregister_kernel_client(sysclient);
		sysclient = -1;
	}
}
